#include <cuda_runtime.h>
#include <cuda_egl_interop.h>
#include <EGL/egl.h>
#include <EGL/eglext.h>
#include <string>
#include <iostream>
#include <sstream>
#include <vector>


static void errorHandler(cudaError_t error, const char* file, int line)
{
    if (error != cudaSuccess) {
        const auto name = cudaGetErrorName(error);
        const auto description = cudaGetErrorString(error);
        std::ostringstream str;
        str << "[CUDA] " << name << " (" << error << "): " << description << " in " << file << ":" << line;
        throw std::runtime_error{ str.str() };
    }
}

static void errorHandler(EGLint error, const char* file, int line)
{
    if (error != EGL_SUCCESS) {
        std::ostringstream str;
        str << "[EGL]  " << error << ". " << file << ":" << line;
        throw std::runtime_error{ str.str() }; 
    }
}

#define checkThrow(x) errorHandler((x), __FILE__, __LINE__)



static PFNEGLCREATESTREAMKHRPROC eglCreateStreamKHR = (PFNEGLCREATESTREAMKHRPROC) eglGetProcAddress("eglCreateStreamKHR");
static PFNEGLDESTROYSTREAMKHRPROC eglDestroyStreamKHR = (PFNEGLDESTROYSTREAMKHRPROC) eglGetProcAddress("eglDestroyStreamKHR");
static PFNEGLSTREAMATTRIBKHRPROC eglStreamAttribKHR = (PFNEGLSTREAMATTRIBKHRPROC) eglGetProcAddress("eglStreamAttribKHR");
static PFNEGLQUERYSTREAMKHRPROC eglQueryStreamKHR = (PFNEGLQUERYSTREAMKHRPROC) eglGetProcAddress("eglQueryStreamKHR");



int main(int argc, char** argv)
{
    size_t WIDTH = 1024;
    size_t HEIGHT = 1024;

    // initialize EGL
    EGLDisplay display = eglGetDisplay(EGL_DEFAULT_DISPLAY);
    checkThrow(eglGetError());
    eglInitialize(display, nullptr, nullptr);
    checkThrow(eglGetError());

    // create mailbox stream
    std::vector<EGLint> mailbox_mode = { EGL_SUPPORT_REUSE_NV, EGL_FALSE, EGL_NONE };
    EGLStreamKHR egl_stream = eglCreateStreamKHR(display, mailbox_mode.data());
    checkThrow(eglGetError());

    // create consumer & producer
    cudaEglStreamConnection consumer;
    checkThrow(cudaEGLStreamConsumerConnect(&consumer, egl_stream));
    cudaEglStreamConnection producer;
    checkThrow(cudaEGLStreamProducerConnect(&producer, egl_stream, WIDTH, HEIGHT));

    for (size_t i = 0; i < 3; ++i) {
        // allocate image, buffer is address that is used twice in the stream, third only once
        uint8_t* buffer = nullptr, *third = nullptr, *src = nullptr;
        checkThrow(cudaMalloc(&buffer, WIDTH * HEIGHT));
        if (i == 2) {
            checkThrow(cudaMalloc(&third, WIDTH * HEIGHT));
            src = third;
        } else {
            src = buffer;
        }

        // fill in surface information
        cudaEglFrame producer_frame = {};
        producer_frame.eglColorFormat = cudaEglColorFormatA;
        producer_frame.frameType = cudaEglFrameTypePitch;
        producer_frame.frame.pPitch[0].ptr = src;
        producer_frame.frame.pPitch[0].pitch = WIDTH;
        producer_frame.planeCount = 1;
        producer_frame.planeDesc[0].width = WIDTH;
        producer_frame.planeDesc[0].height = HEIGHT;
        producer_frame.planeDesc[0].depth = 0;
        producer_frame.planeDesc[0].pitch = WIDTH;
        producer_frame.planeDesc[0].numChannels = 1;
        producer_frame.planeDesc[0].channelDesc = cudaCreateChannelDesc(8, 0, 0, 0, cudaChannelFormatKindUnsigned);

        // present & acquire
        checkThrow(cudaEGLStreamProducerPresentFrame(&producer, producer_frame, nullptr));

        cudaGraphicsResource_t resource = nullptr;
        checkThrow(cudaEGLStreamConsumerAcquireFrame(&consumer, &resource, nullptr, 0));
        cudaEglFrame consumer_frame;
        checkThrow(cudaGraphicsResourceGetMappedEglFrame(&consumer_frame, resource, 0, 0));

        // set all data to 0xff
        checkThrow(cudaMemset(consumer_frame.frame.pPitch[0].ptr, 255, WIDTH * HEIGHT));

        // release & return
        checkThrow(cudaEGLStreamConsumerReleaseFrame(&consumer, resource, nullptr));
        checkThrow(cudaEGLStreamProducerReturnFrame(&producer, &producer_frame, nullptr));

        // check first pixel only, MUST be 0xff
        uint8_t value;
        checkThrow(cudaMemcpy(&value, src, 1, cudaMemcpyDefault));
        std::cout << "i = " << i << std::hex << ", address @ " << (void*) src << ": Pixel: 0x" << int(value) << std::dec << std::endl;

        // free image
        cudaFree(buffer);
        cudaFree(third);
    }

    checkThrow(cudaEGLStreamProducerDisconnect(&producer));
    checkThrow(cudaEGLStreamConsumerDisconnect(&consumer));
    eglDestroyStreamKHR(display, egl_stream);
    eglTerminate(display);

    return 0;
}